/*
 * Copyright (c) Oona Räisänen
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
#ifndef DSP_LIQUID_WRAPPERS_H_
#define DSP_LIQUID_WRAPPERS_H_

#include <array>
#include <complex>
#include <vector>

#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdeprecated-declarations"
// https://github.com/jgaeddert/liquid-dsp/issues/229
#pragma clang diagnostic ignored "-Wreturn-type-c-linkage"
extern "C" {
#include "liquid/liquid.h"
}
#pragma clang diagnostic pop

namespace redsea {

// Hertz to radians per sample
inline float angularFreq(float hertz, float samplerate) {
  return hertz * 2.f * static_cast<float>(M_PI) / samplerate;
}

template <typename T>
struct Maybe {
  T data;
  bool valid;
};

}  // namespace redsea

namespace liquid {

class AGC {
 public:
  AGC(float bw, float initial_gain);
  AGC(const AGC&)             = delete;
  AGC& operator=(const AGC&)  = delete;
  AGC(AGC&& other)            = delete;
  AGC& operator=(AGC&& other) = delete;

  ~AGC();
  std::complex<float> execute(std::complex<float> s);

 private:
  agc_crcf object_;
};

class FIRFilter {
 public:
  FIRFilter(unsigned int len, float fc, float As = 60.0f, float mu = 0.0f);
  FIRFilter(const FIRFilter&)             = delete;
  FIRFilter& operator=(const FIRFilter&)  = delete;
  FIRFilter(FIRFilter&& other)            = delete;
  FIRFilter& operator=(FIRFilter&& other) = delete;

  ~FIRFilter();
  void push(std::complex<float> s);
  std::complex<float> execute();
  size_t length() const;

 private:
  firfilt_crcf object_;
};

class NCO {
 public:
  NCO(liquid_ncotype type, float freq);
  NCO(const NCO&)             = delete;
  NCO& operator=(const NCO&)  = delete;
  NCO(NCO&& other)            = delete;
  NCO& operator=(NCO&& other) = delete;
  ~NCO();
  std::complex<float> mixDown(std::complex<float> s);
  void step();
  void setPLLBandwidth(float);
  void stepPLL(float dphi);
  void reset();

 private:
  nco_crcf object_;
  float initial_frequency_;
};

class SymSync {
 public:
  SymSync(liquid_firfilt_type ftype, unsigned k, unsigned m, float beta, unsigned num_filters);
  SymSync(const SymSync&)             = delete;
  SymSync& operator=(const SymSync&)  = delete;
  SymSync(SymSync&& other)            = delete;
  SymSync& operator=(SymSync&& other) = delete;
  ~SymSync();
  void reset();
  void setBandwidth(float);
  void setOutputRate(unsigned);
  redsea::Maybe<std::complex<float>> execute(std::complex<float>& in);

 private:
  symsync_crcf object_;
  std::vector<std::complex<float>> out_;
};

class Modem {
 public:
  explicit Modem(modulation_scheme scheme);
  Modem(const Modem&)             = delete;
  Modem& operator=(const Modem&)  = delete;
  Modem(Modem&& other)            = delete;
  Modem& operator=(Modem&& other) = delete;
  ~Modem();
  unsigned int demodulate(std::complex<float> sample);
  float getPhaseError();

 private:
#ifdef MODEM_IS_MODEMCF
  modemcf object_;
#else
  modem object_;
#endif
};

class Resampler {
 public:
  static constexpr std::size_t kOutputArraySize{2ULL};

  explicit Resampler(unsigned int length);
  Resampler(const Resampler&)             = delete;
  Resampler& operator=(const Resampler&)  = delete;
  Resampler(Resampler&& other)            = delete;
  Resampler& operator=(Resampler&& other) = delete;
  ~Resampler();
  void setRatio(float ratio);

  unsigned int execute(float in, std::array<float, kOutputArraySize>& out);

 private:
  resamp_rrrf object_;
};

}  // namespace liquid

#endif  // DSP_LIQUID_WRAPPERS_H_
